#[repr(u8)]pub enum MeasurementType {
Range = 0,
Doppler = 1,
Azimuth = 2,
Elevation = 3,
ReceiveFrequency = 4,
TransmitFrequency = 5,
TransmitFrequencyRate = 9,
X = 6,
Y = 7,
Z = 8,
}Variants
Range = 0
Doppler = 1
Azimuth = 2
Elevation = 3
ReceiveFrequency = 4
TransmitFrequency = 5
TransmitFrequencyRate = 9
X = 6
Y = 7
Z = 8
Implementations
Sourceimpl MeasurementType
impl MeasurementType
Sourcepub fn to_field(&self) -> Field
pub fn to_field(&self) -> Field
Returns the fields for this kind of measurement. The metadata includes a unit field with the unit.
Column is nullable in case there is no such measurement at a given epoch.
Sourcepub fn compute_one_way(self, aer: AzElRange, noise: f64) -> Result<f64, ODError>
pub fn compute_one_way(self, aer: AzElRange, noise: f64) -> Result<f64, ODError>
Computes the one way measurement from an AER object and the noise of this measurement type, returned in the units of this measurement type.
Sourcepub fn compute_two_way(
self,
aer_t0: AzElRange,
aer_t1: AzElRange,
noise: f64,
) -> Result<f64, ODError>
pub fn compute_two_way( self, aer_t0: AzElRange, aer_t1: AzElRange, noise: f64, ) -> Result<f64, ODError>
Computes the two way measurement from two AER values and the noise of this measurement type, returned in the units of this measurement type. Two way is modeled by averaging the measurement in between both times, and adding the noise divided by sqrt(2).
Sourcepub fn ccsds_tdm_name(&self) -> &str
pub fn ccsds_tdm_name(&self) -> &str
Returns the CCSDS TDM name for this measurement type.