TrackingDataArc

Struct TrackingDataArc 

pub struct TrackingDataArc {
    pub measurements: BTreeMap<Epoch, Measurement>,
    pub source: Option<String>,
    pub moduli: Option<IndexMap<MeasurementType, f64>>,
    pub force_reject: bool,
}
Tracking data storing all of measurements as a B-Tree. It inherently does NOT support multiple concurrent measurements from several trackers.

Measurement Moduli, e.g. range modulus

In the case of ranging, and possibly other data types, a code is used to measure the range to the spacecraft. The length of this code determines the ambiguity resolution, as per equation 9 in section 2.2.2.2 of the JPL DESCANSO, document 214, Pseudo-Noise and Regenerative Ranging. For example, using the JPL Range Code and a frequency range clock of 1 MHz, the range ambiguity is 75,660 km. In other words, as soon as the spacecraft is at a range of 75,660 + 1 km the JPL Range Code will report the vehicle to be at a range of 1 km. This is simply because the range code overlaps with itself, effectively loosing track of its own reference: it’s due to the phase shift of the signal “lapping” the original signal length.

            (Spacecraft)
            ^
            |    Actual Distance = 75,661 km
            |
0 km                                         75,660 km (Wrap-Around)
|-----------------------------------------------|
  When the "code length" is exceeded,
  measurements wrap back to 0.

So effectively:
    Observed code range = Actual range (mod 75,660 km)
    75,661 km → 1 km

Nyx can only resolve the range ambiguity if the tracking data specifies a modulus for this specific measurement type. For example, in the case of the JPL Range Code and a 1 MHz range clock, the ambiguity interval is 75,660 km.

The measurement used in the Orbit Determination Process then becomes the following, where // represents the Euclidian division.

k = computed_obs // ambiguity_interval
real_obs = measured_obs + k * modulus

Reference: JPL DESCANSO, document 214, Pseudo-Noise and Regenerative Ranging.

Fields

measurements: BTreeMap<Epoch, Measurement>

All measurements in this data arc

source: Option<String>

Source file if loaded from a file or saved to a file.

moduli: Option<IndexMap<MeasurementType, f64>>

Optionally provide a map of modulos (e.g. the RANGE_MODULO of CCSDS TDM).

force_reject: bool

Reject all of the measurements, useful for debugging passes.

Implementations

Source

impl TrackingDataArc

pub fn from_tdm<P: AsRef<Path>>( path: P, aliases: Option<HashMap<String, String>>, ) -> Result<Self, InputOutputError>

Loads a tracking arc from its serialization in CCSDS TDM.

Support level
  • Only the KVN format is supported.
  • Support is limited to orbit determination in “xGEO”, i.e. cislunar and deep space missions.
  • Only one metadata and data section per file is tested.
Data types

Fully supported: - RANGE - DOPPLER_INSTANTANEOUS, DOPPLER_INTEGRATED - ANGLE_1 / ANGLE_2, as azimuth/elevation only

Partially supported: - TRANSMIT_FREQ / RECEIVE_FREQ : these will be converted to Doppler measurements using the TURNAROUND_NUMERATOR and TURNAROUND_DENOMINATOR in the TDM. The freq rate is not supported.

Metadata support
Mode

Only the MODE = SEQUENTIAL is supported.

Time systems / time scales

All timescales supported by hifitime are supported here. This includes: UTC, TAI, GPS, TT, TDB, TAI, GST, QZSST, TL, TCL.

Path

Only one way or two way data is supported, i.e. path must be either PATH n,m,n or PATH n,m.

Note that the actual indexes of the path are ignored.

Participants

PARTICIPANT_1 must be the ground station / tracker. The second participant is ignored: the user must ensure that the Orbit Determination Process is properly configured and the proper arc is given.

Turnaround ratio

The turnaround ratio is only accounted for when the data contains RECEIVE_FREQ and TRANSMIT_FREQ data.

Range and modulus

Only kilometers are supported in range units. Range modulus is accounted for to compute range ambiguity.

pub fn to_tdm_file<P: AsRef<Path>>( self, spacecraft_name: String, aliases: Option<HashMap<String, String>>, path: P, cfg: ExportCfg, ) -> Result<PathBuf, InputOutputError>

Store this tracking arc to a CCSDS TDM file, with optional metadata and a timestamp appended to the filename.

Source

impl TrackingDataArc

pub fn from_parquet<P: AsRef<Path>>(path: P) -> Result<Self, InputOutputError>

Loads a tracking arc from its serialization in parquet.

Warning: no metadata is read from the parquet file, even that written to it by Nyx.

pub fn to_parquet_simple<P: AsRef<Path>>( &self, path: P, ) -> Result<PathBuf, Box<dyn Error>>

Store this tracking arc to a parquet file.

Examples found in repository?
nyx-core/examples/05_cislunar_spacecraft_link_od/main.rs (line 179)
34fn main() -> Result<(), Box<dyn Error>> {
35    pel::init();
36
37    // ====================== //
38    // === ALMANAC SET UP === //
39    // ====================== //
40
41    let manifest_dir = PathBuf::from(env!("CARGO_MANIFEST_DIR"));
42
43    let out = manifest_dir.join("data/04_output/");
44
45    let almanac = Arc::new(
46        Almanac::new(
47            &manifest_dir
48                .join("data/01_planetary/pck08.pca")
49                .to_string_lossy(),
50        )
51        .unwrap()
52        .load(
53            &manifest_dir
54                .join("data/01_planetary/de440s.bsp")
55                .to_string_lossy(),
56        )
57        .unwrap(),
58    );
59
60    let eme2k = almanac.frame_info(EARTH_J2000).unwrap();
61    let moon_iau = almanac.frame_info(IAU_MOON_FRAME).unwrap();
62
63    let epoch = Epoch::from_gregorian_tai(2021, 5, 29, 19, 51, 16, 852_000);
64    let nrho = Orbit::cartesian(
65        166_473.631_302_239_7,
66        -274_715.487_253_382_7,
67        -211_233.210_176_686_7,
68        0.933_451_604_520_018_4,
69        0.436_775_046_841_900_9,
70        -0.082_211_021_250_348_95,
71        epoch,
72        eme2k,
73    );
74
75    let tx_nrho_sc = Spacecraft::from(nrho);
76
77    let state_luna = almanac.transform_to(nrho, MOON_J2000, None).unwrap();
78    println!("Start state (dynamics: Earth, Moon, Sun gravity):\n{state_luna}");
79
80    let bodies = vec![EARTH, SUN];
81    let dynamics = SpacecraftDynamics::new(OrbitalDynamics::point_masses(bodies));
82
83    let setup = Propagator::rk89(
84        dynamics,
85        IntegratorOptions::builder().max_step(0.5.minutes()).build(),
86    );
87
88    /* == Propagate the NRHO vehicle == */
89    let prop_time = 1.1 * state_luna.period().unwrap();
90
91    let (nrho_final, mut tx_traj) = setup
92        .with(tx_nrho_sc, almanac.clone())
93        .for_duration_with_traj(prop_time)
94        .unwrap();
95
96    tx_traj.name = Some("NRHO Tx SC".to_string());
97
98    println!("{tx_traj}");
99
100    /* == Propagate an LLO vehicle == */
101    let llo_orbit =
102        Orbit::try_keplerian_altitude(110.0, 1e-4, 90.0, 0.0, 0.0, 0.0, epoch, moon_iau).unwrap();
103
104    let llo_sc = Spacecraft::builder().orbit(llo_orbit).build();
105
106    let (_, llo_traj) = setup
107        .with(llo_sc, almanac.clone())
108        .until_epoch_with_traj(nrho_final.epoch())
109        .unwrap();
110
111    // Export the subset of the first two hours.
112    llo_traj
113        .clone()
114        .filter_by_offset(..2.hours())
115        .to_parquet_simple(out.join("05_caps_llo_truth.pq"))?;
116
117    /* == Setup the interlink == */
118
119    let mut measurement_types = IndexSet::new();
120    measurement_types.insert(MeasurementType::Range);
121    measurement_types.insert(MeasurementType::Doppler);
122
123    let mut stochastics = IndexMap::new();
124
125    let sa45_csac_allan_dev = 1e-11;
126
127    stochastics.insert(
128        MeasurementType::Range,
129        StochasticNoise::from_hardware_range_km(
130            sa45_csac_allan_dev,
131            10.0.seconds(),
132            link_specific::ChipRate::StandardT4B,
133            link_specific::SN0::Average,
134        ),
135    );
136
137    stochastics.insert(
138        MeasurementType::Doppler,
139        StochasticNoise::from_hardware_doppler_km_s(
140            sa45_csac_allan_dev,
141            10.0.seconds(),
142            link_specific::CarrierFreq::SBand,
143            link_specific::CN0::Average,
144        ),
145    );
146
147    let interlink = InterlinkTxSpacecraft {
148        traj: tx_traj,
149        measurement_types,
150        integration_time: None,
151        timestamp_noise_s: None,
152        ab_corr: Aberration::LT,
153        stochastic_noises: Some(stochastics),
154    };
155
156    // Devices are the transmitter, which is our NRHO vehicle.
157    let mut devices = BTreeMap::new();
158    devices.insert("NRHO Tx SC".to_string(), interlink);
159
160    let mut configs = BTreeMap::new();
161    configs.insert(
162        "NRHO Tx SC".to_string(),
163        TrkConfig::builder()
164            .strands(vec![Strand {
165                start: epoch,
166                end: nrho_final.epoch(),
167            }])
168            .build(),
169    );
170
171    let mut trk_sim =
172        TrackingArcSim::with_seed(devices.clone(), llo_traj.clone(), configs, 0).unwrap();
173    println!("{trk_sim}");
174
175    let trk_data = trk_sim.generate_measurements(almanac.clone()).unwrap();
176    println!("{trk_data}");
177
178    trk_data
179        .to_parquet_simple(out.clone().join("nrho_interlink_msr.pq"))
180        .unwrap();
181
182    // Run a truth OD where we estimate the LLO position
183    let llo_uncertainty = SpacecraftUncertainty::builder()
184        .nominal(llo_sc)
185        .x_km(1.0)
186        .y_km(1.0)
187        .z_km(1.0)
188        .vx_km_s(1e-3)
189        .vy_km_s(1e-3)
190        .vz_km_s(1e-3)
191        .build();
192
193    let mut proc_devices = devices.clone();
194
195    // Define the initial estimate, randomized, seed for reproducibility
196    let mut initial_estimate = llo_uncertainty.to_estimate_randomized(Some(0)).unwrap();
197    // Inflate the covariance -- https://github.com/nyx-space/nyx/issues/339
198    initial_estimate.covar *= 2.5;
199
200    // Increase the noise in the devices to accept more measurements.
201
202    for link in proc_devices.values_mut() {
203        for noise in &mut link.stochastic_noises.as_mut().unwrap().values_mut() {
204            *noise.white_noise.as_mut().unwrap() *= 3.0;
205        }
206    }
207
208    let init_err = initial_estimate
209        .orbital_state()
210        .ric_difference(&llo_orbit)
211        .unwrap();
212
213    println!("initial estimate:\n{initial_estimate}");
214    println!("RIC errors = {init_err}",);
215
216    let odp = InterlinkKalmanOD::new(
217        setup.clone(),
218        KalmanVariant::ReferenceUpdate,
219        Some(ResidRejectCrit::default()),
220        proc_devices,
221        almanac.clone(),
222    );
223
224    // Shrink the data to process.
225    let arc = trk_data.filter_by_offset(..2.hours());
226
227    let od_sol = odp.process_arc(initial_estimate, &arc).unwrap();
228
229    println!("{od_sol}");
230
231    od_sol
232        .to_parquet(
233            out.join("05_caps_interlink_od_sol.pq"),
234            ExportCfg::default(),
235        )
236        .unwrap();
237
238    let od_traj = od_sol.to_traj().unwrap();
239
240    od_traj
241        .ric_diff_to_parquet(
242            &llo_traj,
243            out.join("05_caps_interlink_llo_est_error.pq"),
244            ExportCfg::default(),
245        )
246        .unwrap();
247
248    let final_est = od_sol.estimates.last().unwrap();
249    assert!(final_est.within_3sigma(), "should be within 3 sigma");
250
251    println!("ESTIMATE\n{final_est:x}\n");
252    let truth = llo_traj.at(final_est.epoch()).unwrap();
253    println!("TRUTH\n{truth:x}");
254
255    let final_err = truth
256        .orbit
257        .ric_difference(&final_est.orbital_state())
258        .unwrap();
259    println!("ERROR {final_err}");
260
261    // Build the residuals versus reference plot.
262    let rvr_sol = odp
263        .process_arc(initial_estimate, &arc.resid_vs_ref_check())
264        .unwrap();
265
266    rvr_sol
267        .to_parquet(
268            out.join("05_caps_interlink_resid_v_ref.pq"),
269            ExportCfg::default(),
270        )
271        .unwrap();
272
273    let final_rvr = rvr_sol.estimates.last().unwrap();
274
275    println!("RMAG error {:.3} m", final_err.rmag_km() * 1e3);
276    println!(
277        "Pure prop error {:.3} m",
278        final_rvr
279            .orbital_state()
280            .ric_difference(&final_est.orbital_state())
281            .unwrap()
282            .rmag_km()
283            * 1e3
284    );
285
286    Ok(())
287}
More examples
Hide additional examples
nyx-core/examples/06_lunar_orbit_determination/main.rs (line 158)
35fn main() -> Result<(), Box<dyn Error>> {
36    pel::init();
37
38    // ====================== //
39    // === ALMANAC SET UP === //
40    // ====================== //
41
42    // Dynamics models require planetary constants and ephemerides to be defined.
43    // Let's start by grabbing those by using ANISE's MetaAlmanac.
44
45    let data_folder: PathBuf = [
46        env!("CARGO_MANIFEST_DIR"),
47        "examples",
48        "06_lunar_orbit_determination",
49    ]
50    .iter()
51    .collect();
52
53    let meta = data_folder.join("metaalmanac.dhall");
54
55    // Load this ephem in the general Almanac we're using for this analysis.
56    let almanac = MetaAlmanac::new(meta.to_string_lossy().as_ref())
57        .map_err(Box::new)?
58        .process(true)
59        .map_err(Box::new)?;
60
61    // Lock the almanac (an Arc is a read only structure).
62    let almanac = Arc::new(almanac);
63
64    // Build a nominal trajectory
65    // TODO: Switch this to a sequence once the OD over a spacecraft sequence is implemented.
66
67    let epoch = Epoch::from_gregorian_utc_at_noon(2024, 2, 29);
68    let moon_j2000 = almanac.frame_info(MOON_J2000)?;
69
70    // To build the trajectory we need to provide a spacecraft template.
71    let orbiter = Spacecraft::builder()
72        .mass(Mass::from_dry_and_prop_masses(1018.0, 900.0))
73        .srp(SRPData {
74            area_m2: 3.9 * 2.7,
75            coeff_reflectivity: 0.96,
76        })
77        .orbit(Orbit::try_keplerian_altitude(
78            150.0, 0.00212, 33.6, 45.0, 45.0, 0.0, epoch, moon_j2000,
79        )?) // Setting a zero orbit here because it's just a template
80        .build();
81
82    // ========================== //
83    // === BUILD NOMINAL TRAJ === //
84    // ========================== //
85
86    // Set up the spacecraft dynamics.
87
88    // Specify that the orbital dynamics must account for the graviational pull of the Earth and the Sun.
89    // The gravity of the Moon will also be accounted for since the spaceraft in a lunar orbit.
90    let mut orbital_dyn = OrbitalDynamics::point_masses(vec![EARTH, SUN, JUPITER_BARYCENTER]);
91
92    // We want to include the spherical harmonics, so let's download the gravitational data from the Nyx Cloud.
93    // We're using the GRAIL JGGRX model.
94    let mut jggrx_meta = MetaFile {
95        uri: "http://public-data.nyxspace.com/nyx/models/Luna_jggrx_1500e_sha.tab.gz".to_string(),
96        crc32: Some(0x6bcacda8), // Specifying the CRC32 avoids redownloading it if it's cached.
97    };
98    // And let's download it if we don't have it yet.
99    jggrx_meta.process(true)?;
100
101    // Build the spherical harmonics.
102    // The harmonics must be computed in the body fixed frame.
103    // We're using the long term prediction of the Moon principal axes frame.
104    let moon_pa_frame = MOON_PA_FRAME.with_orient(31008);
105    let sph_harmonics = GravityField::from_stor(
106        almanac.frame_info(moon_pa_frame)?,
107        GravityFieldData::from_shadr(&jggrx_meta.uri, 80, 80, true)?,
108    );
109
110    // Include the spherical harmonics into the orbital dynamics.
111    orbital_dyn.accel_models.push(sph_harmonics);
112
113    // We define the solar radiation pressure, using the default solar flux and accounting only
114    // for the eclipsing caused by the Earth and Moon.
115    // Note that by default, enabling the SolarPressure model will also enable the estimation of the coefficient of reflectivity.
116    let srp_dyn = SolarPressure::new(vec![MOON_J2000], almanac.clone())?;
117
118    // Finalize setting up the dynamics, specifying the force models (orbital_dyn) separately from the
119    // acceleration models (SRP in this case). Use `from_models` to specify multiple accel models.
120    let dynamics = SpacecraftDynamics::from_model(orbital_dyn, srp_dyn);
121
122    println!("{dynamics}");
123
124    let setup = Propagator::rk89(dynamics.clone(), IntegratorOptions::default());
125
126    let truth_traj = setup
127        .with(orbiter, almanac.clone())
128        .for_duration_with_traj(Unit::Day * 2)?
129        .1;
130
131    // ==================== //
132    // === OD SIMULATOR === //
133    // ==================== //
134
135    // Load the Deep Space Network ground stations.
136    // Nyx allows you to build these at runtime but it's pretty static so we can just load them from YAML.
137    let ground_station_file = data_folder.join("dsn-network.yaml");
138    let devices = GroundStation::load_named(ground_station_file)?;
139
140    let proc_devices = devices.clone();
141
142    // Typical OD software requires that you specify your own tracking schedule or you'll have overlapping measurements.
143    // Nyx can build a tracking schedule for you based on the first station with access.
144    let configs: BTreeMap<String, TrkConfig> =
145        TrkConfig::load_named(data_folder.join("tracking-cfg.yaml"))?;
146
147    // Build the tracking arc simulation to generate a "standard measurement".
148    let mut trk = TrackingArcSim::<Spacecraft, GroundStation>::with_seed(
149        devices.clone(),
150        truth_traj.clone(),
151        configs,
152        123, // Set a seed for reproducibility
153    )?;
154
155    trk.build_schedule(almanac.clone())?;
156    let arc = trk.generate_measurements(almanac.clone())?;
157    // Save the simulated tracking data
158    arc.to_parquet_simple("./data/04_output/06_lunar_simulated_tracking.parquet")?;
159
160    // We'll note that in our case, we have continuous coverage of LRO when the vehicle is not behind the Moon.
161    println!("{arc}");
162
163    // Now that we have simulated measurements, we'll run the orbit determination.
164
165    // ===================== //
166    // === OD ESTIMATION === //
167    // ===================== //
168
169    let sc = SpacecraftUncertainty::builder()
170        .nominal(orbiter)
171        .frame(LocalFrame::RIC)
172        .x_km(0.5)
173        .y_km(0.5)
174        .z_km(0.5)
175        .vx_km_s(5e-3)
176        .vy_km_s(5e-3)
177        .vz_km_s(5e-3)
178        .build();
179
180    // Build the filter initial estimate, which we will reuse in the filter.
181    let initial_estimate = sc.to_estimate()?;
182
183    println!("== FILTER STATE ==\n{orbiter:x}\n{initial_estimate}");
184
185    // Build the SNC in the Moon J2000 frame, specified as a velocity noise over time.
186    let process_noise = ProcessNoise3D::from_velocity_km_s(
187        &[1e-14, 1e-14, 1e-14],
188        1 * Unit::Hour,
189        10 * Unit::Minute,
190        None,
191    );
192
193    println!("{process_noise}");
194
195    // We'll set up the OD process to reject measurements whose residuals are move than 3 sigmas away from what we expect.
196    let odp = SpacecraftKalmanScalarOD::new(
197        setup,
198        KalmanVariant::ReferenceUpdate,
199        Some(ResidRejectCrit::default()),
200        proc_devices,
201        almanac.clone(),
202    )
203    .with_process_noise(process_noise);
204
205    let od_sol = odp.process_arc(initial_estimate, &arc)?;
206
207    let final_est = od_sol.estimates.last().unwrap();
208
209    println!("{final_est}");
210
211    let ric_err = truth_traj
212        .at(final_est.epoch())?
213        .orbit
214        .ric_difference(&final_est.orbital_state())?;
215    println!("== RIC at end ==");
216    println!("RIC Position (m): {:.3}", ric_err.radius_km * 1e3);
217    println!("RIC Velocity (m/s): {:.3}", ric_err.velocity_km_s * 1e3);
218
219    println!(
220        "Num residuals rejected: #{}",
221        od_sol.rejected_residuals().len()
222    );
223    println!(
224        "Percentage within +/-3: {}",
225        od_sol.residual_ratio_within_threshold(3.0).unwrap()
226    );
227    println!("Whitened residuals normal? {}", od_sol.is_normal(None)?);
228    println!("NIS test success? {}", od_sol.is_nis_consistent(None)?);
229
230    od_sol.to_parquet(
231        "./data/04_output/06_lunar_od_results.parquet",
232        ExportCfg::default(),
233    )?;
234
235    let od_trajectory = od_sol.to_traj()?;
236    // Build the RIC difference.
237    od_trajectory.ric_diff_to_parquet(
238        &truth_traj,
239        "./data/04_output/06_lunar_od_truth_error.parquet",
240        ExportCfg::default(),
241    )?;
242
243    Ok(())
244}
nyx-core/examples/04_lro_od/main.rs (line 243)
35fn main() -> Result<(), Box<dyn Error>> {
36    pel::init();
37
38    // ====================== //
39    // === ALMANAC SET UP === //
40    // ====================== //
41
42    // Dynamics models require planetary constants and ephemerides to be defined.
43    // Let's start by grabbing those by using ANISE's MetaAlmanac.
44
45    let data_folder: PathBuf = [env!("CARGO_MANIFEST_DIR"), "examples", "04_lro_od"]
46        .iter()
47        .collect();
48
49    let meta = data_folder.join("lro-dynamics.dhall");
50
51    // Load this ephem in the general Almanac we're using for this analysis.
52    let mut almanac = MetaAlmanac::new(meta.to_string_lossy().as_ref())
53        .map_err(Box::new)?
54        .process(true)
55        .map_err(Box::new)?;
56
57    let mut moon_pc = almanac.get_planetary_data_from_id(MOON).unwrap();
58    moon_pc.mu_km3_s2 = 4902.74987;
59    almanac.set_planetary_data_from_id(MOON, moon_pc).unwrap();
60
61    let mut earth = almanac.get_planetary_data_from_id(EARTH).unwrap();
62    earth.mu_km3_s2 = 398600.436;
63    almanac.set_planetary_data_from_id(EARTH, earth).unwrap();
64
65    // Save this new kernel for reuse.
66    // In an operational context, this would be part of the "Lock" process, and should not change throughout the mission.
67    almanac
68        .planetary_data
69        .values()
70        .next()
71        .unwrap()
72        .save_as(&data_folder.join("lro-specific.pca"), true)?;
73
74    // Lock the almanac (an Arc is a read only structure).
75    let almanac = Arc::new(almanac);
76
77    // Orbit determination requires a Trajectory structure, which can be saved as parquet file.
78    // In our case, the trajectory comes from the BSP file, so we need to build a Trajectory from the almanac directly.
79    // To query the Almanac, we need to build the LRO frame in the J2000 orientation in our case.
80    // Inspecting the LRO BSP in the ANISE GUI shows us that NASA has assigned ID -85 to LRO.
81    let lro_frame = Frame::from_ephem_j2000(-85);
82
83    // To build the trajectory we need to provide a spacecraft template.
84    let sc_template = Spacecraft::builder()
85        .mass(Mass::from_dry_and_prop_masses(1018.0, 900.0)) // Launch masses
86        .srp(SRPData {
87            // SRP configuration is arbitrary, but we will be estimating it anyway.
88            area_m2: 3.9 * 2.7,
89            coeff_reflectivity: 0.96,
90        })
91        .orbit(Orbit::zero(MOON_J2000)) // Setting a zero orbit here because it's just a template
92        .build();
93    // Now we can build the trajectory from the BSP file.
94    // We'll arbitrarily set the tracking arc to 24 hours with a five second time step.
95    let traj_as_flown = Traj::from_bsp(
96        lro_frame,
97        MOON_J2000,
98        almanac.clone(),
99        sc_template,
100        5.seconds(),
101        Some(Epoch::from_str("2024-01-01 00:00:00 UTC")?),
102        Some(Epoch::from_str("2024-01-02 00:00:00 UTC")?),
103        Aberration::LT,
104        Some("LRO".to_string()),
105    )?;
106
107    println!("{traj_as_flown}");
108
109    // ====================== //
110    // === MODEL MATCHING === //
111    // ====================== //
112
113    // Set up the spacecraft dynamics.
114
115    // Specify that the orbital dynamics must account for the graviational pull of the Earth and the Sun.
116    // The gravity of the Moon will also be accounted for since the spaceraft in a lunar orbit.
117    let mut orbital_dyn = OrbitalDynamics::point_masses(vec![EARTH, SUN, JUPITER_BARYCENTER]);
118
119    // We want to include the spherical harmonics, so let's download the gravitational data from the Nyx Cloud.
120    // We're using the GRAIL JGGRX model.
121    let mut jggrx_meta = MetaFile {
122        uri: "http://public-data.nyxspace.com/nyx/models/Luna_jggrx_1500e_sha.tab.gz".to_string(),
123        crc32: Some(0x6bcacda8), // Specifying the CRC32 avoids redownloading it if it's cached.
124    };
125    // And let's download it if we don't have it yet.
126    jggrx_meta.process(true)?;
127
128    // Build the spherical harmonics.
129    // The harmonics must be computed in the body fixed frame.
130    // We're using the long term prediction of the Moon principal axes frame.
131    let moon_pa_frame = MOON_PA_FRAME.with_orient(31008);
132    let sph_harmonics = GravityField::from_stor(
133        almanac.frame_info(moon_pa_frame)?,
134        GravityFieldData::from_shadr(&jggrx_meta.uri, 80, 80, true)?,
135    );
136
137    // Include the spherical harmonics into the orbital dynamics.
138    orbital_dyn.accel_models.push(sph_harmonics);
139
140    // We define the solar radiation pressure, using the default solar flux and accounting only
141    // for the eclipsing caused by the Earth and Moon.
142    // Note that by default, enabling the SolarPressure model will also enable the estimation of the coefficient of reflectivity.
143    let srp_dyn = SolarPressure::new(vec![EARTH_J2000, MOON_J2000], almanac.clone())?;
144
145    // Finalize setting up the dynamics, specifying the force models (orbital_dyn) separately from the
146    // acceleration models (SRP in this case). Use `from_models` to specify multiple accel models.
147    let dynamics = SpacecraftDynamics::from_model(orbital_dyn, srp_dyn);
148
149    println!("{dynamics}");
150
151    // Now we can build the propagator.
152    let setup = Propagator::default_dp78(dynamics.clone());
153
154    // For reference, let's build the trajectory with Nyx's models from that LRO state.
155    let (sim_final, traj_as_sim) = setup
156        .with(*traj_as_flown.first(), almanac.clone())
157        .until_epoch_with_traj(traj_as_flown.last().epoch())?;
158
159    println!("SIM INIT:  {:x}", traj_as_flown.first());
160    println!("SIM FINAL: {sim_final:x}");
161    // Compute RIC difference between SIM and LRO ephem
162    let sim_lro_delta = sim_final
163        .orbit
164        .ric_difference(&traj_as_flown.last().orbit)?;
165    println!("{traj_as_sim}");
166    println!(
167        "SIM v LRO - RIC Position (m): {:.3}",
168        sim_lro_delta.radius_km * 1e3
169    );
170    println!(
171        "SIM v LRO - RIC Velocity (m/s): {:.3}",
172        sim_lro_delta.velocity_km_s * 1e3
173    );
174
175    traj_as_sim.ric_diff_to_parquet(
176        &traj_as_flown,
177        "./data/04_output/04_lro_sim_truth_error.parquet",
178        ExportCfg::default(),
179    )?;
180
181    // ==================== //
182    // === OD SIMULATOR === //
183    // ==================== //
184
185    // After quite some time trying to exactly match the model, we still end up with an oscillatory difference on the order of 150 meters between the propagated state
186    // and the truth LRO state.
187
188    // Therefore, we will actually run an estimation from a dispersed LRO state.
189    // The sc_seed is the true LRO state from the BSP.
190    let sc_seed = *traj_as_flown.first();
191
192    // Load the Deep Space Network ground stations.
193    // Nyx allows you to build these at runtime but it's pretty static so we can just load them from YAML.
194    let ground_station_file: PathBuf = [
195        env!("CARGO_MANIFEST_DIR"),
196        "examples",
197        "04_lro_od",
198        "dsn-network.yaml",
199    ]
200    .iter()
201    .collect();
202
203    let devices = GroundStation::load_named(ground_station_file)?;
204
205    let mut proc_devices = devices.clone();
206
207    // Increase the noise in the devices to accept more measurements.
208    for gs in proc_devices.values_mut() {
209        if let Some(noise) = &mut gs
210            .stochastic_noises
211            .as_mut()
212            .unwrap()
213            .get_mut(&MeasurementType::Range)
214        {
215            *noise.white_noise.as_mut().unwrap() *= 3.0;
216        }
217    }
218
219    // Typical OD software requires that you specify your own tracking schedule or you'll have overlapping measurements.
220    // Nyx can build a tracking schedule for you based on the first station with access.
221    let trkconfg_yaml: PathBuf = [
222        env!("CARGO_MANIFEST_DIR"),
223        "examples",
224        "04_lro_od",
225        "tracking-cfg.yaml",
226    ]
227    .iter()
228    .collect();
229
230    let configs: BTreeMap<String, TrkConfig> = TrkConfig::load_named(trkconfg_yaml)?;
231
232    // Build the tracking arc simulation to generate a "standard measurement".
233    let mut trk = TrackingArcSim::<Spacecraft, GroundStation>::with_seed(
234        devices.clone(),
235        traj_as_flown.clone(),
236        configs,
237        123, // Set a seed for reproducibility
238    )?;
239
240    trk.build_schedule(almanac.clone())?;
241    let arc = trk.generate_measurements(almanac.clone())?;
242    // Save the simulated tracking data
243    arc.to_parquet_simple("./data/04_output/04_lro_simulated_tracking.parquet")?;
244
245    // We'll note that in our case, we have continuous coverage of LRO when the vehicle is not behind the Moon.
246    println!("{arc}");
247
248    // Now that we have simulated measurements, we'll run the orbit determination.
249
250    // ===================== //
251    // === OD ESTIMATION === //
252    // ===================== //
253
254    let sc = SpacecraftUncertainty::builder()
255        .nominal(sc_seed)
256        .frame(LocalFrame::RIC)
257        .x_km(0.5)
258        .y_km(0.5)
259        .z_km(0.5)
260        .vx_km_s(5e-3)
261        .vy_km_s(5e-3)
262        .vz_km_s(5e-3)
263        .build();
264
265    // Build the filter initial estimate, which we will reuse in the filter.
266    let mut initial_estimate = sc.to_estimate()?;
267    initial_estimate.covar *= 3.0;
268
269    println!("== FILTER STATE ==\n{sc_seed:x}\n{initial_estimate}");
270
271    // Build the SNC in the Moon J2000 frame, specified as a velocity noise over time.
272    let process_noise = ProcessNoise3D::from_velocity_km_s(
273        &[1e-12, 1e-12, 1e-12],
274        1 * Unit::Hour,
275        10 * Unit::Minute,
276        None,
277    );
278
279    println!("{process_noise}");
280
281    // We'll set up the OD process to reject measurements whose residuals are move than 3 sigmas away from what we expect.
282    let odp = SpacecraftKalmanOD::new(
283        setup,
284        KalmanVariant::ReferenceUpdate,
285        Some(ResidRejectCrit::default()),
286        proc_devices,
287        almanac.clone(),
288    )
289    .with_process_noise(process_noise);
290
291    let od_sol = odp.process_arc(initial_estimate, &arc)?;
292
293    let final_est = od_sol.estimates.last().unwrap();
294
295    println!("{final_est}");
296
297    let ric_err = traj_as_flown
298        .at(final_est.epoch())?
299        .orbit
300        .ric_difference(&final_est.orbital_state())?;
301    println!("== RIC at end ==");
302    println!("RIC Position (m): {:.3}", ric_err.radius_km * 1e3);
303    println!("RIC Velocity (m/s): {:.3}", ric_err.velocity_km_s * 1e3);
304
305    println!(
306        "Num residuals rejected: #{}",
307        od_sol.rejected_residuals().len()
308    );
309    println!(
310        "Percentage within +/-3: {}",
311        od_sol.residual_ratio_within_threshold(3.0).unwrap()
312    );
313    println!("Ratios normal? {}", od_sol.is_normal(None).unwrap());
314
315    od_sol.to_parquet(
316        "./data/04_output/04_lro_od_results.parquet",
317        ExportCfg::default(),
318    )?;
319
320    // Create the ephemeris
321    let ephem = od_sol.to_ephemeris("LRO rebuilt".to_string());
322    let ephem_start = ephem.start_epoch().unwrap();
323    let ephem_end = ephem.end_epoch().unwrap();
324    // Check that the covariance is PSD throughout the ephemeris by interpolating it.
325    for epoch in TimeSeries::inclusive(ephem_start, ephem_end, Unit::Minute * 5) {
326        ephem
327            .covar_at(
328                epoch,
329                anise::ephemerides::ephemeris::LocalFrame::RIC,
330                &almanac,
331            )
332            .unwrap_or_else(|e| panic!("covar not PSD at {epoch}: {e}"));
333    }
334    // Export as BSP!
335    ephem
336        .write_spice_bsp(-85, "./data/04_output/04_lro_rebuilt.bsp", None)
337        .expect("could not built BSP");
338    let new_almanac = Almanac::default()
339        .load("./data/04_output/04_lro_rebuilt.bsp")
340        .unwrap();
341    new_almanac.describe(None, None, None, None, None, None, None, None);
342    let (spk_start, spk_end) = new_almanac.spk_domain(-85).unwrap();
343
344    assert!((ephem_start - spk_start).abs() < Unit::Microsecond * 1);
345    assert!((ephem_end - spk_end).abs() < Unit::Microsecond * 1);
346
347    // In our case, we have the truth trajectory from NASA.
348    // So we can compute the RIC state difference between the real LRO ephem and what we've just estimated.
349    // Export the OD trajectory first.
350    let od_trajectory = od_sol.to_traj()?;
351    // Build the RIC difference.
352    od_trajectory.ric_diff_to_parquet(
353        &traj_as_flown,
354        "./data/04_output/04_lro_od_truth_error.parquet",
355        ExportCfg::default(),
356    )?;
357
358    Ok(())
359}
pub fn to_parquet<P: AsRef<Path>>( &self, path: P, cfg: ExportCfg, ) -> Result<PathBuf, Box<dyn Error>>

Store this tracking arc to a parquet file, with optional metadata and a timestamp appended to the filename.

Source

impl TrackingDataArc

pub fn start_epoch(&self) -> Option<Epoch>

Returns the start epoch of this tracking arc

pub fn end_epoch(&self) -> Option<Epoch>

Returns the end epoch of this tracking arc

pub fn duration(&self) -> Option<Duration>

Returns the duration this tracking arc

pub fn len(&self) -> usize

Returns the number of measurements in this data arc

pub fn is_empty(&self) -> bool

Returns whether this arc has no measurements.

pub fn min_duration_sep(&self) -> Option<Duration>

Returns the minimum duration between two subsequent measurements.

Source

impl TrackingDataArc

pub fn set_moduli(&mut self, msr_type: MeasurementType, modulus: f64)

Set (or overwrites) the modulus of the provided measurement type.

pub fn apply_moduli(&mut self)

Applies the moduli to each measurement, if defined.

pub fn unique_aliases(&self) -> IndexSet<String>

Returns the unique list of aliases in this tracking data arc

pub fn unique_types(&self) -> IndexSet<MeasurementType>

Returns the unique measurement types in this tracking data arc

pub fn unique(&self) -> (IndexSet<String>, IndexSet<MeasurementType>)

Returns the unique trackers and unique measurement types in this data arc

pub fn filter_by_epoch<R: RangeBounds<Epoch>>(self, bound: R) -> Self

Returns a new tracking arc that only contains measurements that fall within the given epoch range.

pub fn filter_by_offset<R: RangeBounds<Duration>>(self, bound: R) -> Self

Returns a new tracking arc that only contains measurements that fall within the given offset from the first epoch. For example, a bound of 30.minutes()..90.minutes() will only read measurements from the start of the arc + 30 minutes until start + 90 minutes.

Examples found in repository?
nyx-core/examples/05_cislunar_spacecraft_link_od/main.rs (line 225)
34fn main() -> Result<(), Box<dyn Error>> {
35    pel::init();
36
37    // ====================== //
38    // === ALMANAC SET UP === //
39    // ====================== //
40
41    let manifest_dir = PathBuf::from(env!("CARGO_MANIFEST_DIR"));
42
43    let out = manifest_dir.join("data/04_output/");
44
45    let almanac = Arc::new(
46        Almanac::new(
47            &manifest_dir
48                .join("data/01_planetary/pck08.pca")
49                .to_string_lossy(),
50        )
51        .unwrap()
52        .load(
53            &manifest_dir
54                .join("data/01_planetary/de440s.bsp")
55                .to_string_lossy(),
56        )
57        .unwrap(),
58    );
59
60    let eme2k = almanac.frame_info(EARTH_J2000).unwrap();
61    let moon_iau = almanac.frame_info(IAU_MOON_FRAME).unwrap();
62
63    let epoch = Epoch::from_gregorian_tai(2021, 5, 29, 19, 51, 16, 852_000);
64    let nrho = Orbit::cartesian(
65        166_473.631_302_239_7,
66        -274_715.487_253_382_7,
67        -211_233.210_176_686_7,
68        0.933_451_604_520_018_4,
69        0.436_775_046_841_900_9,
70        -0.082_211_021_250_348_95,
71        epoch,
72        eme2k,
73    );
74
75    let tx_nrho_sc = Spacecraft::from(nrho);
76
77    let state_luna = almanac.transform_to(nrho, MOON_J2000, None).unwrap();
78    println!("Start state (dynamics: Earth, Moon, Sun gravity):\n{state_luna}");
79
80    let bodies = vec![EARTH, SUN];
81    let dynamics = SpacecraftDynamics::new(OrbitalDynamics::point_masses(bodies));
82
83    let setup = Propagator::rk89(
84        dynamics,
85        IntegratorOptions::builder().max_step(0.5.minutes()).build(),
86    );
87
88    /* == Propagate the NRHO vehicle == */
89    let prop_time = 1.1 * state_luna.period().unwrap();
90
91    let (nrho_final, mut tx_traj) = setup
92        .with(tx_nrho_sc, almanac.clone())
93        .for_duration_with_traj(prop_time)
94        .unwrap();
95
96    tx_traj.name = Some("NRHO Tx SC".to_string());
97
98    println!("{tx_traj}");
99
100    /* == Propagate an LLO vehicle == */
101    let llo_orbit =
102        Orbit::try_keplerian_altitude(110.0, 1e-4, 90.0, 0.0, 0.0, 0.0, epoch, moon_iau).unwrap();
103
104    let llo_sc = Spacecraft::builder().orbit(llo_orbit).build();
105
106    let (_, llo_traj) = setup
107        .with(llo_sc, almanac.clone())
108        .until_epoch_with_traj(nrho_final.epoch())
109        .unwrap();
110
111    // Export the subset of the first two hours.
112    llo_traj
113        .clone()
114        .filter_by_offset(..2.hours())
115        .to_parquet_simple(out.join("05_caps_llo_truth.pq"))?;
116
117    /* == Setup the interlink == */
118
119    let mut measurement_types = IndexSet::new();
120    measurement_types.insert(MeasurementType::Range);
121    measurement_types.insert(MeasurementType::Doppler);
122
123    let mut stochastics = IndexMap::new();
124
125    let sa45_csac_allan_dev = 1e-11;
126
127    stochastics.insert(
128        MeasurementType::Range,
129        StochasticNoise::from_hardware_range_km(
130            sa45_csac_allan_dev,
131            10.0.seconds(),
132            link_specific::ChipRate::StandardT4B,
133            link_specific::SN0::Average,
134        ),
135    );
136
137    stochastics.insert(
138        MeasurementType::Doppler,
139        StochasticNoise::from_hardware_doppler_km_s(
140            sa45_csac_allan_dev,
141            10.0.seconds(),
142            link_specific::CarrierFreq::SBand,
143            link_specific::CN0::Average,
144        ),
145    );
146
147    let interlink = InterlinkTxSpacecraft {
148        traj: tx_traj,
149        measurement_types,
150        integration_time: None,
151        timestamp_noise_s: None,
152        ab_corr: Aberration::LT,
153        stochastic_noises: Some(stochastics),
154    };
155
156    // Devices are the transmitter, which is our NRHO vehicle.
157    let mut devices = BTreeMap::new();
158    devices.insert("NRHO Tx SC".to_string(), interlink);
159
160    let mut configs = BTreeMap::new();
161    configs.insert(
162        "NRHO Tx SC".to_string(),
163        TrkConfig::builder()
164            .strands(vec![Strand {
165                start: epoch,
166                end: nrho_final.epoch(),
167            }])
168            .build(),
169    );
170
171    let mut trk_sim =
172        TrackingArcSim::with_seed(devices.clone(), llo_traj.clone(), configs, 0).unwrap();
173    println!("{trk_sim}");
174
175    let trk_data = trk_sim.generate_measurements(almanac.clone()).unwrap();
176    println!("{trk_data}");
177
178    trk_data
179        .to_parquet_simple(out.clone().join("nrho_interlink_msr.pq"))
180        .unwrap();
181
182    // Run a truth OD where we estimate the LLO position
183    let llo_uncertainty = SpacecraftUncertainty::builder()
184        .nominal(llo_sc)
185        .x_km(1.0)
186        .y_km(1.0)
187        .z_km(1.0)
188        .vx_km_s(1e-3)
189        .vy_km_s(1e-3)
190        .vz_km_s(1e-3)
191        .build();
192
193    let mut proc_devices = devices.clone();
194
195    // Define the initial estimate, randomized, seed for reproducibility
196    let mut initial_estimate = llo_uncertainty.to_estimate_randomized(Some(0)).unwrap();
197    // Inflate the covariance -- https://github.com/nyx-space/nyx/issues/339
198    initial_estimate.covar *= 2.5;
199
200    // Increase the noise in the devices to accept more measurements.
201
202    for link in proc_devices.values_mut() {
203        for noise in &mut link.stochastic_noises.as_mut().unwrap().values_mut() {
204            *noise.white_noise.as_mut().unwrap() *= 3.0;
205        }
206    }
207
208    let init_err = initial_estimate
209        .orbital_state()
210        .ric_difference(&llo_orbit)
211        .unwrap();
212
213    println!("initial estimate:\n{initial_estimate}");
214    println!("RIC errors = {init_err}",);
215
216    let odp = InterlinkKalmanOD::new(
217        setup.clone(),
218        KalmanVariant::ReferenceUpdate,
219        Some(ResidRejectCrit::default()),
220        proc_devices,
221        almanac.clone(),
222    );
223
224    // Shrink the data to process.
225    let arc = trk_data.filter_by_offset(..2.hours());
226
227    let od_sol = odp.process_arc(initial_estimate, &arc).unwrap();
228
229    println!("{od_sol}");
230
231    od_sol
232        .to_parquet(
233            out.join("05_caps_interlink_od_sol.pq"),
234            ExportCfg::default(),
235        )
236        .unwrap();
237
238    let od_traj = od_sol.to_traj().unwrap();
239
240    od_traj
241        .ric_diff_to_parquet(
242            &llo_traj,
243            out.join("05_caps_interlink_llo_est_error.pq"),
244            ExportCfg::default(),
245        )
246        .unwrap();
247
248    let final_est = od_sol.estimates.last().unwrap();
249    assert!(final_est.within_3sigma(), "should be within 3 sigma");
250
251    println!("ESTIMATE\n{final_est:x}\n");
252    let truth = llo_traj.at(final_est.epoch()).unwrap();
253    println!("TRUTH\n{truth:x}");
254
255    let final_err = truth
256        .orbit
257        .ric_difference(&final_est.orbital_state())
258        .unwrap();
259    println!("ERROR {final_err}");
260
261    // Build the residuals versus reference plot.
262    let rvr_sol = odp
263        .process_arc(initial_estimate, &arc.resid_vs_ref_check())
264        .unwrap();
265
266    rvr_sol
267        .to_parquet(
268            out.join("05_caps_interlink_resid_v_ref.pq"),
269            ExportCfg::default(),
270        )
271        .unwrap();
272
273    let final_rvr = rvr_sol.estimates.last().unwrap();
274
275    println!("RMAG error {:.3} m", final_err.rmag_km() * 1e3);
276    println!(
277        "Pure prop error {:.3} m",
278        final_rvr
279            .orbital_state()
280            .ric_difference(&final_est.orbital_state())
281            .unwrap()
282            .rmag_km()
283            * 1e3
284    );
285
286    Ok(())
287}
pub fn filter_by_tracker(self, tracker: String) -> Self

Returns a new tracking arc that only contains measurements from the desired tracker.

pub fn filter_by_measurement_type(self, included_type: MeasurementType) -> Self

Returns a new tracking arc that only contains measurements of the provided type.

pub fn exclude_tracker(self, excluded_tracker: String) -> Self

Returns a new tracking arc that contains measurements from all trackers except the one provided

pub fn exclude_by_epoch<R: RangeBounds<Epoch>>(self, bound: R) -> Self

Returns a new tracking arc that excludes measurements within the given epoch range.

pub fn exclude_measurement_type(self, excluded_type: MeasurementType) -> Self

Returns a new tracking arc that contains measurements from all trackers except the one provided

pub fn reject_by_epoch<R: RangeBounds<Epoch>>(self, bound: R) -> Self

Marks measurements within the given epoch range as rejected.

pub fn reject_by_tracker(self, tracker: String) -> Self

Marks measurements from the provided tracker as rejected.

pub fn downsample(self, target_step: Duration) -> Self

Downsamples the tracking data to a lower frequency using a simple moving average low-pass filter followed by decimation, returning new TrackingDataArc with downsampled measurements.

It provides a computationally efficient approach to reduce the sampling rate while mitigating aliasing effects.

Algorithm
  1. A simple moving average filter is applied as a low-pass filter.
  2. Decimation is performed by selecting every Nth sample after filtering.
Advantages
  • Computationally efficient, suitable for large datasets common in spaceflight applications.
  • Provides basic anti-aliasing, crucial for preserving signal integrity in orbit determination and tracking.
  • Maintains phase information, important for accurate timing in spacecraft state estimation.
Limitations
  • The frequency response is not as sharp as more sophisticated filters (e.g., FIR, IIR).
  • May not provide optimal stopband attenuation for high-precision applications.
Considerations for Spaceflight Applications
  • Suitable for initial data reduction in ground station tracking pipelines.
  • Adequate for many orbit determination and tracking tasks where computational speed is prioritized.
  • For high-precision applications (e.g., interplanetary navigation), consider using more advanced filtering techniques.
pub fn resid_vs_ref_check(self) -> Self

Examples found in repository?
nyx-core/examples/05_cislunar_spacecraft_link_od/main.rs (line 263)
34fn main() -> Result<(), Box<dyn Error>> {
35    pel::init();
36
37    // ====================== //
38    // === ALMANAC SET UP === //
39    // ====================== //
40
41    let manifest_dir = PathBuf::from(env!("CARGO_MANIFEST_DIR"));
42
43    let out = manifest_dir.join("data/04_output/");
44
45    let almanac = Arc::new(
46        Almanac::new(
47            &manifest_dir
48                .join("data/01_planetary/pck08.pca")
49                .to_string_lossy(),
50        )
51        .unwrap()
52        .load(
53            &manifest_dir
54                .join("data/01_planetary/de440s.bsp")
55                .to_string_lossy(),
56        )
57        .unwrap(),
58    );
59
60    let eme2k = almanac.frame_info(EARTH_J2000).unwrap();
61    let moon_iau = almanac.frame_info(IAU_MOON_FRAME).unwrap();
62
63    let epoch = Epoch::from_gregorian_tai(2021, 5, 29, 19, 51, 16, 852_000);
64    let nrho = Orbit::cartesian(
65        166_473.631_302_239_7,
66        -274_715.487_253_382_7,
67        -211_233.210_176_686_7,
68        0.933_451_604_520_018_4,
69        0.436_775_046_841_900_9,
70        -0.082_211_021_250_348_95,
71        epoch,
72        eme2k,
73    );
74
75    let tx_nrho_sc = Spacecraft::from(nrho);
76
77    let state_luna = almanac.transform_to(nrho, MOON_J2000, None).unwrap();
78    println!("Start state (dynamics: Earth, Moon, Sun gravity):\n{state_luna}");
79
80    let bodies = vec![EARTH, SUN];
81    let dynamics = SpacecraftDynamics::new(OrbitalDynamics::point_masses(bodies));
82
83    let setup = Propagator::rk89(
84        dynamics,
85        IntegratorOptions::builder().max_step(0.5.minutes()).build(),
86    );
87
88    /* == Propagate the NRHO vehicle == */
89    let prop_time = 1.1 * state_luna.period().unwrap();
90
91    let (nrho_final, mut tx_traj) = setup
92        .with(tx_nrho_sc, almanac.clone())
93        .for_duration_with_traj(prop_time)
94        .unwrap();
95
96    tx_traj.name = Some("NRHO Tx SC".to_string());
97
98    println!("{tx_traj}");
99
100    /* == Propagate an LLO vehicle == */
101    let llo_orbit =
102        Orbit::try_keplerian_altitude(110.0, 1e-4, 90.0, 0.0, 0.0, 0.0, epoch, moon_iau).unwrap();
103
104    let llo_sc = Spacecraft::builder().orbit(llo_orbit).build();
105
106    let (_, llo_traj) = setup
107        .with(llo_sc, almanac.clone())
108        .until_epoch_with_traj(nrho_final.epoch())
109        .unwrap();
110
111    // Export the subset of the first two hours.
112    llo_traj
113        .clone()
114        .filter_by_offset(..2.hours())
115        .to_parquet_simple(out.join("05_caps_llo_truth.pq"))?;
116
117    /* == Setup the interlink == */
118
119    let mut measurement_types = IndexSet::new();
120    measurement_types.insert(MeasurementType::Range);
121    measurement_types.insert(MeasurementType::Doppler);
122
123    let mut stochastics = IndexMap::new();
124
125    let sa45_csac_allan_dev = 1e-11;
126
127    stochastics.insert(
128        MeasurementType::Range,
129        StochasticNoise::from_hardware_range_km(
130            sa45_csac_allan_dev,
131            10.0.seconds(),
132            link_specific::ChipRate::StandardT4B,
133            link_specific::SN0::Average,
134        ),
135    );
136
137    stochastics.insert(
138        MeasurementType::Doppler,
139        StochasticNoise::from_hardware_doppler_km_s(
140            sa45_csac_allan_dev,
141            10.0.seconds(),
142            link_specific::CarrierFreq::SBand,
143            link_specific::CN0::Average,
144        ),
145    );
146
147    let interlink = InterlinkTxSpacecraft {
148        traj: tx_traj,
149        measurement_types,
150        integration_time: None,
151        timestamp_noise_s: None,
152        ab_corr: Aberration::LT,
153        stochastic_noises: Some(stochastics),
154    };
155
156    // Devices are the transmitter, which is our NRHO vehicle.
157    let mut devices = BTreeMap::new();
158    devices.insert("NRHO Tx SC".to_string(), interlink);
159
160    let mut configs = BTreeMap::new();
161    configs.insert(
162        "NRHO Tx SC".to_string(),
163        TrkConfig::builder()
164            .strands(vec![Strand {
165                start: epoch,
166                end: nrho_final.epoch(),
167            }])
168            .build(),
169    );
170
171    let mut trk_sim =
172        TrackingArcSim::with_seed(devices.clone(), llo_traj.clone(), configs, 0).unwrap();
173    println!("{trk_sim}");
174
175    let trk_data = trk_sim.generate_measurements(almanac.clone()).unwrap();
176    println!("{trk_data}");
177
178    trk_data
179        .to_parquet_simple(out.clone().join("nrho_interlink_msr.pq"))
180        .unwrap();
181
182    // Run a truth OD where we estimate the LLO position
183    let llo_uncertainty = SpacecraftUncertainty::builder()
184        .nominal(llo_sc)
185        .x_km(1.0)
186        .y_km(1.0)
187        .z_km(1.0)
188        .vx_km_s(1e-3)
189        .vy_km_s(1e-3)
190        .vz_km_s(1e-3)
191        .build();
192
193    let mut proc_devices = devices.clone();
194
195    // Define the initial estimate, randomized, seed for reproducibility
196    let mut initial_estimate = llo_uncertainty.to_estimate_randomized(Some(0)).unwrap();
197    // Inflate the covariance -- https://github.com/nyx-space/nyx/issues/339
198    initial_estimate.covar *= 2.5;
199
200    // Increase the noise in the devices to accept more measurements.
201
202    for link in proc_devices.values_mut() {
203        for noise in &mut link.stochastic_noises.as_mut().unwrap().values_mut() {
204            *noise.white_noise.as_mut().unwrap() *= 3.0;
205        }
206    }
207
208    let init_err = initial_estimate
209        .orbital_state()
210        .ric_difference(&llo_orbit)
211        .unwrap();
212
213    println!("initial estimate:\n{initial_estimate}");
214    println!("RIC errors = {init_err}",);
215
216    let odp = InterlinkKalmanOD::new(
217        setup.clone(),
218        KalmanVariant::ReferenceUpdate,
219        Some(ResidRejectCrit::default()),
220        proc_devices,
221        almanac.clone(),
222    );
223
224    // Shrink the data to process.
225    let arc = trk_data.filter_by_offset(..2.hours());
226
227    let od_sol = odp.process_arc(initial_estimate, &arc).unwrap();
228
229    println!("{od_sol}");
230
231    od_sol
232        .to_parquet(
233            out.join("05_caps_interlink_od_sol.pq"),
234            ExportCfg::default(),
235        )
236        .unwrap();
237
238    let od_traj = od_sol.to_traj().unwrap();
239
240    od_traj
241        .ric_diff_to_parquet(
242            &llo_traj,
243            out.join("05_caps_interlink_llo_est_error.pq"),
244            ExportCfg::default(),
245        )
246        .unwrap();
247
248    let final_est = od_sol.estimates.last().unwrap();
249    assert!(final_est.within_3sigma(), "should be within 3 sigma");
250
251    println!("ESTIMATE\n{final_est:x}\n");
252    let truth = llo_traj.at(final_est.epoch()).unwrap();
253    println!("TRUTH\n{truth:x}");
254
255    let final_err = truth
256        .orbit
257        .ric_difference(&final_est.orbital_state())
258        .unwrap();
259    println!("ERROR {final_err}");
260
261    // Build the residuals versus reference plot.
262    let rvr_sol = odp
263        .process_arc(initial_estimate, &arc.resid_vs_ref_check())
264        .unwrap();
265
266    rvr_sol
267        .to_parquet(
268            out.join("05_caps_interlink_resid_v_ref.pq"),
269            ExportCfg::default(),
270        )
271        .unwrap();
272
273    let final_rvr = rvr_sol.estimates.last().unwrap();
274
275    println!("RMAG error {:.3} m", final_err.rmag_km() * 1e3);
276    println!(
277        "Pure prop error {:.3} m",
278        final_rvr
279            .orbital_state()
280            .ric_difference(&final_est.orbital_state())
281            .unwrap()
282            .rmag_km()
283            * 1e3
284    );
285
286    Ok(())
287}
pub fn chunk(&self, max_duration: Duration) -> Vec<TrackingDataArc>

Splits a long tracking data arc into smaller chunks, each up to max_duration long. This is inspired by JPL MONTE’s long arc setup to ensure BLSE convergence on manageable chunks.